# # @Author: Felix Kramer <kramer>
# # @Date:   23-06-2021
# # @Email:  kramer@mpi-cbg.de
# # @Project: phd_network_remodelling
# @Last modified by:   felix
# @Last modified time: 2022-07-02T13:51:06+02:00

import numpy as np
from dataclasses import dataclass, field
# custom
from .base import model


@dataclass
class meigel(model):
    """
    The network adaptation model according to Meigel/Horton et al, based on
    minimization of supply-demand mismatch, with a simplified metabolite uptake
    model though:
        Meigel et al, Flow rate of transport network controls uniform
        metabolite supply to tissue, Journal of the Royal Society Interface,
        2018\n
        Kramer&Modes, On biological flow networks: Antagonism between
        hydrodynamic and metabolic stimuli as driver of topological
        transitions, arXiv:2111.04326
    The system's cost function is given as:\n
    .. math::
        \\Gamma = \\sum_{e} \\left[ \\Phi_e-\\Phi_0 \\right]^2
    Attributes:
        pars (dict):\n
         The specific model parameters \\alpha_0, \\alpha_1 and demand \\Phi_0

        ivp_options (dict):\n
         Information to generate the internal solver_options. Providing t0,t1
         number of evaluation and x0.\n
        model_args (list):\n
         Model specific paramerets need to evaluate the update rules of the DS
         \n
        solver_options (dict):\n
         Specifying runtime and evaluation marks. \n
        events (dict):\n
         Events to consier, here in gernal flatlining events which allow for
         early termination of stiff simulations. \n
        null_decimals (in):\n
         description \n

    Methods:
        init()
            The model post_init function.
        update_event_func()
            Update the event function and ensures that solver_options are set.
        flatlining_default(t, x_0, *args)
            The default flatlining function for determining the terminal event
            of the dynamical system.
        flatlining_dynamic(t, x_0, *args)
            The dynamic flatlining function for determining the terminal event
            of the dynamical system.
        set_model_parameters(model_pars)
            Set internal model arguments array.
        set_solver_options(solv_opt)
            Set internal solver_options and update event function.
        calc_update_stimuli(t, x_0, *args)
            The dynamic system's temporal update rule, computing the gradient
            -dF for dx/dt.
        calc_cost_stimuli(t, x_0, *args)
            Computes the dynamic system's Lyapunov function and its gradient.
        get_stimuli_pars(self, flow, x_0)
            Update flow & pressure landscapes, recompute conductivity as well
            as squared quantities.

    """

    pars: dict = field(default_factory=dict, init=True, repr=True)

    def __post_init__(self):

        self.init()

        self.solver_options.update({'events': 'default'})
        if self.pars:
            self.set_model_parameters(self.pars)
        else:
            self.model_args = [1.]

    def calc_update_stimuli(self, t, x_0, flux, phi0):

        m = len(x_0)
        dx = np.zeros(m)
        # c, B_new = flux.calc_profile_concentration()
        flux.update_transport_matrix(x_0)
        flux.solve_absorbing_boundary()

        phi = flux.calc_absorption()
        J_phi = flux.calc_absorption_jacobian()
        phi0 = np.ones(m)*phi0

        for i in range(m):
            dx[i] = -2.*np.sum(
                    np.multiply(
                        np.subtract(phi, phi0),
                        J_phi[i, :])
                    )

        return dx

    def calc_cost_stimuli(self, t, x_0, flux, phi0):

        m = len(x_0)
        cnd, p_sq = self.get_stimuli_pars(flux, x_0)
        flux.circuit.C = cnd[:]
        flux.circuit.R = x_0[:]

        # c, B_new = flux.calc_profile_concentration()
        flux.update_transport_matrix(x_0)
        flux.solve_absorbing_boundary()
        phi = flux.calc_absorption()
        phi0 = np.ones(m)*phi0
        F = np.sum(np.power(np.subtract(phi, phi0), 2))

        # J_phi = flux.calc_absorption_jacobian()
        # dF = np.array([ 2.*np.sum(np.multiply(
        # np.subtract(phi, phi0), J_phi[j, :])) for j in range(m)])
        dF = -self.calc_update_stimuli(t, x_0, flux, phi0)

        return F, dF

    def calc_cost_system(self, t, x_0, flux, phi0):

        m = len(x_0)
        cnd, p_sq = self.get_stimuli_pars(flux, x_0)

        flux.circuit.C = cnd[:]
        flux.circuit.R = x_0[:]

        flux.update_transport_matrix(x_0)
        phi = flux.calc_absorption()
        phi0 = np.ones(m)*phi0
        F = np.sum(np.power(np.subtract(phi, phi0), 2))

        return F

    def get_stimuli_pars(self, flux, x_0):

        k = flux.circuit.scales['conductance']
        src = flux.circuit.nodes['source']

        x_sq = np.power(x_0, 2)
        cnd = flux.calc_conductivity_from_cross_section(x_sq, k)
        p_sq, q_sq = flux.calc_sq_flow(cnd, src)

        return cnd, p_sq


@dataclass
class link(meigel):
    """
    The network adaptation model according to Meigel/Horton et al, based on
    flow dissipation-volume minimization in combination with supply-demand
    competition of conductivity costs as published for example:
        Kramer&Modes, On biological flow networks: Antagonism between
        hydrodynamic and metabolic stimuli as driver of topological
        transitions, arXiv:2111.04326

    The system's cost function is given as:\n
    .. math::
        \\Gamma = \\sum_{e}\\left( \\alpha_1 r_e^4\\Delta p_e^2 +
        \\alpha_0 r_e^2 + \\left[ \\Phi_e-\\Phi_0 \\right]^2 \\right)
    Attributes:
        pars (dict):\n
         The specific model parameters \\alpha_0, \\alpha_1 and demand \\Phi_0

        ivp_options (dict):\n
         Information to generate the internal solver_options. Providing t0,t1
         number of evaluation and x0.\n
        model_args (list):\n
         Model specific paramerets need to evaluate the update rules of the DS
         \n
        solver_options (dict):\n
         Specifying runtime and evaluation marks. \n
        events (dict):\n
         Events to consier, here in gernal flatlining events which allow for
         early termination of stiff simulations. \n
        null_decimals (in):\n
         description \n

    Methods:
        init()
            The model post_init function.
        update_event_func()
            Update the event function and ensures that solver_options are set.
        flatlining_default(t, x_0, *args)
            The default flatlining function for determining the terminal event
            of the dynamical system.
        flatlining_dynamic(t, x_0, *args)
            The dynamic flatlining function for determining the terminal event
            of the dynamical system.
        set_model_parameters(model_pars)
            Set internal model arguments array.
        set_solver_options(solv_opt)
            Set internal solver_options and update event function.
        calc_update_stimuli(t, x_0, *args)
            The dynamic system's temporal update rule, computing the gradient
            -dF for dx/dt.
        calc_cost_stimuli(t, x_0, *args)
            Computes the dynamic system's Lyapunov function and its gradient.
        get_stimuli_pars(self, flow, x_0)
            Update flow & pressure landscapes, recompute conductivity as well
            as squared quantities.

    """

    def __post_init__(self):

        self.init()

        self.solver_options.update({'events': 'default'})
        if self.pars:
            self.set_model_parameters(self.pars)
        else:
            self.model_args = [1., 1., 1.]

    def calc_update_stimuli(self, t, x_0, flux, phi0, alpha_0, alpha_1):

        m = len(x_0)

        # calc absorption stimulus
        dx1 = np.zeros(m)

        flux.update_transport_matrix(x_0)
        flux.solve_absorbing_boundary()

        phi = flux.calc_absorption()
        J_phi = flux.calc_absorption_jacobian()
        phi0 = np.ones(m)*phi0

        for i in range(m):
            dx1[i] = -2.*np.sum(
                    np.multiply(
                        np.subtract(phi, phi0),
                        J_phi[i, :])
                )

        # calc hydrodynamic stimulus
        r_sq = np.power(x_0, 2)
        k = flux.circuit.scales['conductance']
        src = flux.circuit.nodes['source']
        conductivity = flux.calc_conductivity_from_cross_section(r_sq, k)
        dP, P = flux.calc_pressure(conductivity, src)

        s1 = 2.*alpha_1*np.multiply(np.power(dP, 2), r_sq)
        s2 = alpha_0*np.ones(m)
        ds = np.subtract(s1, s2)
        dx2 = 2*np.multiply(ds, x_0)

        return np.add(dx1, dx2)

    def calc_cost_stimuli(self, t, x_0, flux, phi0, alpha_0, alpha_1):

        m = len(x_0)
        cnd, p_sq = self.get_stimuli_pars(flux, x_0)
        flux.circuit.C = cnd[:]
        flux.circuit.R = x_0[:]

        # c, B_new = flux.calc_profile_concentration()
        flux.update_transport_matrix(x_0)
        flux.solve_absorbing_boundary()
        phi = flux.calc_absorption()
        phi0 = np.ones(m)*phi0
        F = np.sum(np.power(np.subtract(phi, phi0), 2))

        f1 = alpha_1*np.multiply(p_sq, np.power(x_0, 4))
        f2 = alpha_0*np.power(x_0, 2)
        c2 = np.sum(np.add(f1, f2))

        F += c2
        # J_phi = flux.calc_absorption_jacobian()
        # dF = np.array([ 2.*np.sum(
        # np.multiply(np.subtract(phi, phi0), J_phi[j, :])) for j in range(m)])
        dF = -self.calc_update_stimuli(t, x_0, flux, phi0, alpha_0, alpha_1)

        return F, dF

    def calc_cost_system(self, t, x_0, flux, phi0, alpha_0, alpha_1):

        m = len(x_0)
        cnd, p_sq = self.get_stimuli_pars(flux, x_0)

        flux.circuit.C = cnd[:]
        flux.circuit.R = x_0[:]

        flux.update_transport_matrix(x_0)
        phi = flux.calc_absorption()
        phi0 = np.ones(m)*phi0
        F = np.sum(np.power(np.subtract(phi, phi0), 2))

        f1 = alpha_1*np.multiply(p_sq, np.power(x_0, 4))
        f2 = alpha_0*np.power(x_0, 2)
        c2 = np.sum(np.add(f1, f2))
        F += c2

        return F

    def get_stimuli_pars(self, flux, x_0):

        k = flux.circuit.scales['conductance']
        src = flux.circuit.nodes['source']

        x_sq = np.power(x_0, 2)
        cnd = flux.calc_conductivity_from_cross_section(x_sq, k)
        p_sq, q_sq = flux.calc_sq_flow(cnd, src)

        return cnd, p_sq


# @dataclass
# class volume(link):
#
#     def __post_init__(self):
#
#         self.init()
#
#         self.solver_options.update({'events': 'default'})
#         if self.pars:
#             self.set_model_parameters(self.pars)
#         else:
#             self.model_args = [1., 1., 1.]
#
#     # def hexagonal_grid(self, *args):
#     #
#     #     tiling_factor, sidelength, conductance,flow,periodic_bool=args
#     #     self.k=conductance
#     #     self.l=sidelength
#     #     self.f=flow
#     #
#     #     m=2*tiling_factor+1
#     #     n=2*tiling_factor
#     #     self.G=nx.hexagonal_lattice_graph(
#           m, n, periodic=periodic_bool, with_positions=True)
#     #     for n in self.G.nodes():
#     #         self.G.nodes[n]['label']=self.count_n
#     #         self.G.nodes[n]['pos']=self.l*np.array(self.G.nodes[n]['pos'])
#     #     for e in self.G.edges():
#     #
#     #         self.G.edges[e]['label']=self.count_e
#     #         self.G.edges[e]['slope']=[
# self.G.nodes[e[0]]['pos'],
# self.G.nodes[e[1]]['pos']
# ]
#     #
#     #     # initialze circuit & add attributes
#     #     self.initialize_circuit()
#     #
#     # def set_up_volumes(self,periodic_bool):
#     #
#     #     # set up volumes be aware to work on properly iniced systems
#     #     if periodic_bool:
#     #         T=analyze_graph.tool_box()
#     #         cycle_basis=T.construct_minimum_basis(self.G)
#     #         dict_volumes={}
#     #         dict_idx={}
#     #         dict_counter={}
#     #         for i,e in enumerate(self.G.edges()):
#     #             dict_idx[e]=i
#     #             dict_counter[e]=0
#     #         for i,c in enumerate(cycle_basis):
#     #                 dict_volumes[i]=[]
#     #                 for j,e in enumerate( c.edges() ):
#     #
#     #                     if e in dict_idx:
#     #                         dict_volumes[i].append(dict_idx[e])
#     #                         dict_counter[e]+=1
#     #                     else:
#     #                         dict_volumes[i].append(dict_idx[(e[-1],e[0])])
#     #                         dict_counter[(e[-1],e[0])]+=1
#     #         dict_volumes[len(cycle_basis)]=[]
#     #         for i,e in enumerate(self.G.edges()):
#     #             if dict_counter[e]<2:
#     #                 dict_volumes[len(cycle_basis)].append(dict_idx[e])
#     #
#     #         keys=list(dict_volumes.keys())
#     #         for k in keys:
#     #             if len(dict_volumes[k])>6:
#     #                 del dict_volumes[k]
#     #         self.dict_volumes=dict_volumes
#     #     else:
#     #         print('why bother?')
#
#
#     def calc_update_stimuli(self, t, x_0, flux, phi0, alpha_0, alpha_1):
#
#         m = len(x_0)
#         # calc absorption stimulus
#         dx1 = np.zeros(m)
#
#         flux.update_transport_matrix(x_0)
#         flux.solve_absorbing_boundary()
#         phi0 = np.ones(m)*phi0
#         phi = flux.calc_absorption()
#         J_phi = flux.calc_absorption_jacobian()
#         dphi = np.ones(m)
#
#         # FIX
#         # for i, v in enumerate(K.dict_volumes.keys()):
#         #     dphi[i] = np.sum(phi[K.dict_volumes[v]])-phi0[i]
#         #
#         # for j, e in enumerate(self.list_e):
#         #     for i, v in enumerate(K.dict_volumes.keys()):
#         #         dr[j] -=2.*dphi[i]*np.sum(J_phi[j, K.dict_volumes[v]])
#
#         # calc hydrodynamic stimulus
#         r_sq = np.power(x_0, 2)
#         k = flux.circuit.scales['conductance']
#         src = flux.circuit.nodes['source']
#         conductivity = flux.calc_conductivity_from_cross_section(r_sq, k)
#         dP, P = flux.calc_pressure(conductivity,src)
#
#         s1 = 2.*alpha_1*np.multiply( np.power(dP,2), r_sq)
#         s2 = alpha_0*np.ones(m)
#         ds = np.subtract(s1, s2)
#         dx2 = 2*np.multiply(ds, x_0)
#
#         return np.add(dx1,dx2)
#
#     def calc_cost_stimuli(self, t, x_0, flux, phi0, alpha_0, alpha_1):
#
#         m = len(x_0)
#         cnd, p_sq = self.get_stimuli_pars(flux, x_0)
#         flux.circuit.C = cnd[:]
#         flux.circuit.R = x_0[:]
#
#         # c, B_new = flux.calc_profile_concentration()
#         flux.update_transport_matrix(x_0)
#         flux.solve_absorbing_boundary()
#         phi = flux.calc_absorption()
#         phi0 = np.ones(m)*phi0
#         F = np.sum(np.power(np.subtract(phi, phi0), 2))
#
#         f1 = alpha_1*np.multiply(p_sq, np.power(x_0, 4))
#         f2 = alpha_0 *np.power(x_0, 2)
#         c2 = np.sum(np.add(f1, f2))
#
#         F+=c2
#         # J_phi = flux.calc_absorption_jacobian()
#         # dF = np.array([ 2.*np.sum(np.multiply(
# np.subtract(phi, phi0), J_phi[j, :])) for j in range(m)])
#         dF = -self.calc_update_stimuli(t, x_0, flux, phi0, alpha_0, alpha_1)
#
#         return F, dF
#
#     def calc_cost_system(self, t, x_0, flux, phi0):
#
#         m = len(x_0)
#         cnd, p_sq = self.get_stimuli_pars(flux, x_0)
#
#         flux.circuit.C = cnd[:]
#         flux.circuit.R = x_0[:]
#
#         flux.update_transport_matrix(x_0)
#         phi = flux.calc_absorption()
#         phi0 = np.ones(m)*phi0
#         F = np.sum(np.power(np.subtract(phi, phi0), 2))
#
#         f1 = alpha_1*np.multiply(p_sq, np.power(x_0, 4))
#         f2 = alpha_0 *np.power(x_0, 2)
#         c2 = np.sum(np.add(f1, f2))
#         F+=c2
#
#         return F
#
#     def get_stimuli_pars(self, flux, x_0):
#
#         k = flux.circuit.scales['conductance']
#         src = flux.circuit.nodes['source']
#
#         x_sq = np.power(x_0, 2)
#         cnd = flux.calc_conductivity_from_cross_section(x_sq, k)
#         p_sq, q_sq =flux.calc_sq_flow(cnd, src)
#
#         return cnd, p_sq



# @Author: Felix Kramer <felix>
# @Date:   2022-06-29T18:27:48+02:00
# @Email:  felixuwekramer@proton.me
# @Filename: init_ivp.py
# @Last modified by:   felix
# @Last modified time: 2022-07-01T18:56:54+02:00

import sys
import numpy as np
import networkx as nx
# import scipy.optimize as sc
import scipy.integrate as si
from ..models import base, binder
from dataclasses import dataclass, field
# general initial value problem for network morpgogenesis


@dataclass
class proxy_solver():
    """
    A proxy solver class for custom numeric integration.

    Args:
        tsamples (array):\n
         Setting the internal sample time points.
        \n
        sol (array):\n
         Time series of data to be stored internally.
    Returns:
        type: proxy_solver
    """

    defVal = dict(default_factory=list, init=True, repr=False)
    t_samples: list = field(**defVal)
    sol: list = field(**defVal)

    def __post_init__(self):

        self.t = self.t_samples
        self.y = self.sol.transpose()


@dataclass
class morph():
    """
    Basic Class for simulating flow network adaption dynamics.

    'Morph' computes the long-term network development on the basis of
    flow and gradient descent. The class acts as a wrapper to ensure correct
    pre-/post-proccessing of the network's data and output. Utilizing the
    modules 'kirchhoff' and 'hailhydro' as well as goflow.models.

    Args:
        construct (nx.Graph):\n
         A network plexus, recommended to be 'kirchhoff' circuit or an
         initialized 'hailhydro' flow/flux. Handling for deviations only
         provided for standard models.\n
        mode (string/model):\n
         A descriptive string for selection internal model object.\n
        args (tuple):\n
         A parameter tuple for evaluation of the model given a plexus.\n
    Returns:
        type: morph
    """

    defVal = dict(init=True, repr=True)

    construct: nx.Graph = field(**defVal)
    mode: str = field(default='default', **defVal)
    args: tuple = field(default_factory=tuple(), **defVal)

    def __post_init__(self):
        """
        Post object initialization methods.

        Streamline provided model and flow parameters and set dynamics of the
        ODE system.

        Args:
            None

        Returns:
            None
        """

        self.init_model_and_flow()
        self.link_model_flow()

    def link_model_flow(self):
        """
        Transfer model and flow parameters to internal variables of solver.

        Args:
            None

        Returns:
            None
        """

        args = dict(args=(self.flow, *self.model.model_args))
        self.model.solver_options.update(args)
        self.model.update_event_func()

    def init_model_and_flow(self):
        """
        Set internal model and flow variables with constructor values.

        Args:
            None

        Returns:
            None

        Raises:
            Exception: \n
             sys.exit('Terminate Program: No valid adaptation model provided.')

        """

        if isinstance(self.mode, base.model):

            self.flow = self.construct
            self.model = self.mode

        elif self.mode in binder.modelBinder:

            model = binder.modelBinder[self.mode]
            flow = binder.circuitBinder[self.mode]

            self.model = model(self.args[0])
            self.flow = flow(self.construct, *self.args[1:])

        else:
            sys.exit('Terminate Program: No valid adaptation model provided.')


@dataclass
class morph_dynamic(morph):
    """
    Derived class definition from morph, for specific dynamic models.

    Defines explicit ODE solvers and customizable forward integration solvers.

    Args:
        construct (nx.Graph):\n
         A network plexus, recommended to be 'kirchhoff' circuit or an
         initialized 'hailhydro' flow/flux. Handling for deviations only
         provided for standard models.\n
        mode (string/model):\n
         A descriptive string for selection internal model object.\n
        args (tuple):\n
         A parameter tuple for evaluation of the model given a plexus.\n
    Returns:
        type: morph_dynamic

    """

    def __post_init__(self):
        """
        Post object initialization methods.

        Streamline provided model and flow parameters and set dynamics of the
        ODE system.

        Args:
            None

        Returns:
            None
        """
        self.evals = 100
        self.init_model_and_flow()
        self.link_model_flow()

    def autoSolve(self, t_span, x0):
        """
        Default ODE solver/wrapper for given adaptation models.

        Based on methods of scipy.integrate, hence solver options are
        customizable via the internal model.solver_options attribute and align
        with scipy standards.

        Args:
            t_span (tuple):\n
             A tuple setting the linear range of values for
             explicit evaluation.\n
            x0 (array):\n
             An array of initial values to start the numeric integration of the
             dynamic system.\n
        Returns:
            Iterable: Same set of return values as scipy.integrate.solve_ivp
            (1.7.3)

        """

        self.options = {
            # 'method': 'RK45',
            'method': 'LSODA',
            # 'method': 'BDF',
            'atol': 1e-10,
            'rtol': 1e-7,
            # 'dense_output': True,
        }

        for k, v in self.model.solver_options.items():
            self.options[k] = v
        self.options.update({
            't_eval': np.linspace(t_span[0], t_span[1], num=self.evals)
        })

        ds_func = self.model.calc_update_stimuli

        nsol = si.solve_ivp(ds_func, t_span, x0, **self.options)

        return nsol

    def nlogSolve(self, t_span, x0):
        """
        Custom ODE solver/wrapper for given adaptation models with logarithmic
        evaluation timescales.

        Based on methods of scipy.integrate, hence solver options are
        customizable via the internal model.solver_options attribute and align
        with scipy standards.

        Args:
            t_span (tuple):\n
             A tuple setting the logarithmic range of values for
             explicit evaluation.\n
            x0 (array):\n
             An array of initial values to start the numeric
             integration of the dynamic system.\n
        Returns:
            Iterable: Same set of return values as scipy.integrate.solve_ivp
            (1.7.3)

        """

        self.options = {
            # 'method': 'RK45',
            'method': 'LSODA',
            # 'method': 'BDF',
            'atol': 1e-10,
            'rtol': 1e-7,
            # 'dense_output': True,
        }

        for k, v in self.model.solver_options.items():
            self.options[k] = v
        self.options.update({
            't_eval': np.logspace(
                        np.log10(t_span[0]),
                        np.log10(t_span[1]),
                        num=self.evals
                        )
        })

        ds_func = self.model.calc_update_stimuli

        nsol = si.solve_ivp(ds_func, t_span, x0, **self.options)

        return nsol

    def nsolve(self, ds_func, t_span, x0, **kwargs):
        """
        Custom ODE solver/wrapper for given adaptation models with linear
        evaluation timescales and custom update function. Based on methods of
        scipy.integrate.

        Args:
            ds_func (callable function):\n
             The update function of the dynamic system, which is supposed to
             define dx/dt, according to scipy standards.\n
            t_span (tuple):\n
             A tuple setting the logarithmic range of values for
             explicit evaluation.\n
            x0 (array):\n
             An array of initial values to start the numeric
             integration of the dynamic system.\n
            kwargs (dict):\n
             Dictionary for setting solver options according to
             scipy.integrate.solve_ivp standards.\n
        Returns:
            Iterable: Same set of return values as scipy.integrate.solve_ivp
            (1.7.3)
        """

        self.options = {
            'method': 'LSODA',
            'atol': 1e-10,
            'dense_output': True,
        }
        for k, v in kwargs.items():
            self.options[k] = v

        self.options.update({
            't_eval': np.linspace(t_span[0], t_span[1], num=self.evals)
        })

        nsol = si.solve_ivp(ds_func, t_span, x0, **self.options)

        return nsol

    def nsolve_custom(self, ds_func, x0, **kwargs):
        """
        Custom ODE solver onthe basis of Foward-Euler, for given adaptation
        models with custom evaluation timescales and step width and custom
        update function. Acts as a wrapper for nsolve_fw_euler.

        Args:
            ds_func (callable function):\n
             The update function of the dynamic system, which is supposed to
            define dx/dt, according to scipy standards.\n
            t_span (tuple):\n
             A tuple setting the logarithmic range of values for explicit
             evaluation.\n
            x0 (array):\n
             An array of initial values to start the numeric
             integration of the dynamic system.\n
            kwargs (dict):\n
             A dictionary setting custom solver options such as
             number of evaluations 'sample', number of steps 'num_steps'\n
        Returns:
            Iterable: Tuple of two arrays (t, x(t))
        """

        self.options = {
            'num_steps': 1,
            'samples': 1,
            'step': 1,
        }
        for k, v in kwargs.items():
            self.options[k] = v

        ns, sr = self.set_integration_scale(
                        self.options['num_steps'],
                        self.options['samples']
                        )
        self.options['sample_rate'] = sr
        self.options['num_steps'] = ns

        nsol = self.nsolve_fw_euler(ds_func, x0, **self.options)

        return nsol

    def nsolve_fw_euler(self, ds_func, x0, **kwargs):
        """
        Custom ODE solver onthe basis of Foward-Euler, for given adaptation
        models with custom evaluation timescales and step width and custom
        update function.

        Args:
            ds_func (callable):\n
             The update function of the dynamic system, which is supposed to
             define dx/dt, according to scipy standards.\n
            t_span (tuple):\n
             A tuple setting the logarithmic range of values for
             explicit evaluation.\n
            x0 (array):\n
             An array of initial values to start the numeric integration of the
             dynamic system.\n
            kwargs (dict):\n
             A dictionary setting custom solver options.\n
        Returns:
            Iterable: Tuple of two arrays (t, x(t))
        """

        t_samples = kwargs['step']*np.arange(
                        0,
                        kwargs['num_steps'],
                        step=kwargs['sample_rate']
                        )

        sol = np.zeros((kwargs['samples'], len(x0)))
        c_m = 0
        x_0 = np.array(x0)

        for i in range(kwargs['num_steps']):

            if (i % kwargs['sample_rate']) == 0:
                sol[c_m] = x_0[:]
                c_m += 1

            dx = ds_func(i*kwargs['step'], x_0, *kwargs['args'])
            x_0 = np.add(x_0, dx*kwargs['step'])

        nsol = proxy_solver(t_samples, sol)

        return nsol

    def set_integration_scale(self, Num_steps, sample):
        """
        Computes the sample rate for the given dynamic system.

        Given the pre-set number of computing steps and number of evaluations,
        Adjusts/increases the number of integration cycles such that last
        evaluation coincicdes with last cycle.

        Args:
            Num_steps (int):\n
             Number of integration steps to perform for a Forward-Euler scheme.
             \n
            sample(int):\n
             Number of desired function evaluations for output.\n
        Returns:
            Iterable: Num_steps, sample_rate
        """

        # reshape number of integration steps & sample rates for consistency
        sample_rate = int(Num_steps/sample)
        if (sample_rate*sample) < Num_steps:
            Num_steps = sample_rate*sample

        return Num_steps, sample_rate
# TODO: GlobalOptimizers
# @dataclass
# class MySteps():
#
#     stepsize: float = 0.
#     # def __init__(self, stepsize ):
#     #     self.stepsize  = stepsize
#     def __call__(self, x):
#         rx = np.add(x, np.random.rand(len(x))*self.stepsize)
#         return rx

# @dataclass
# class morph_optimize(morph):
#
#     def __post_init__(self):
#         E = self.flow.circuit.list_graph_edges
#         mysteps = MySteps(1.)
#         b0 = 1e-25
#         self.options = {
#             'step': mysteps,
#             'niter': 100,
#             'T': 10.,
#             'minimizer_kwargs': {
#                 'method': 'L-BFGS-B',
#                 'bounds': [(b0, None) for x in range(len(E))],
#                 'args': (self.flow.circuit),
#                 'jac': False,
#                 'tol': 1e-10
#                 }
#         }
#
#     def update_minimizer_options(**kwargs):
#
#         if 'step' in kwargs:
#             mysteps = MySteps(kwargs['step'])
#             kwargs['step'] = mysteps
#
#         for k, v in kwargs.items():
#             if k in self.options:
#                 options[k] = v
#
#         if 'minimizer_kwargs' in kwargs:
#             for ks, vs in kwargs['minimizer_kwargs']:
#                 minimizer_kwargs[ks] = vs
#
#     def optimize_network(self, cost_func, x0, **kwargs):
#
#         update_minimizer_options(**kwargs)
#
#         sol = sc.basinhopping(cost_func, x0, **self.options)
#
#         return sol



# @Author: Felix Kramer <felix>
# @Date:   2022-06-28T16:24:41+02:00
# @Email:  felixuwekramer@proton.me
# @Filename: base.py
# @Last modified by:   felix
# @Last modified time: 2022-07-02T14:16:48+02:00

import numpy as np
from dataclasses import dataclass, field


@dataclass
class model():
    """
    The basic model class, acting as a center piece for 'morph' simulations, as
    it defines the individual dynamic system.

    This class defines the ODE sytem, metabolic cost model and a diverse set of
    functionalities for effective evaluation of any simulation's proceeding.
    Based on scipy.integrate.solve_ivp (1.7.3) it defines specific event
    handlers for stiff systems and utility functions.

    Attributes:
        ivp_options (dict):\n
         Information to generate the internal solver_options. Providing t0,t1
         number of evaluation and x0.\n
        model_args (list):\n
         Model specific paramerets need to evaluate the update rules of the DS
         \n
        solver_options (dict):\n
         Specifying runtime and evaluation marks. \n
        events (dict):\n
         Events to consier, here in gernal flatlining events which allow for
         early termination of stiff simulations. \n
        null_decimals (in):\n
         description \n

    Methods:
        init()
            The model post_init function.
        update_event_func()
            Update the event function and ensures that solver_options are set.
        flatlining_default(t, x_0, *args)
            The default flatlining function for determining the terminal event
            of the dynamical system.
        flatlining_dynamic(t, x_0, *args)
            The dynamic flatlining function for determining the terminal event
            of the dynamical system.
        set_model_parameters(model_pars)
            Set internal model arguments array.
        set_solver_options(solv_opt)
            Set internal solver_options and update event function.
        calc_update_stimuli(t, x_0, *args)
            The dynamic system's temporal update rule, computing the gradient
            -dF for dx/dt.
        calc_cost_stimuli(t, x_0, *args)
            Computes the dynamic system's Lyapunov function and its gradient.

    """

    defVals = dict(init=False, repr=False)

    ivp_options: dict = field(default_factory=dict, **defVals)
    model_args: list = field(default_factory=list, **defVals)
    solver_options: dict = field(default_factory=dict, init=False, repr=True)
    events: dict = field(default_factory=dict, **defVals)
    null_decimal: int = field(default=6, **defVals)
    # jac: bool = field(default=False, **defVals)

    def __post_init__(self):

        self.init()

    def init(self):
        """
        The model post_init function.

        Setting proxy values for internal variables.

        """

        self.ivp_options = {
            't0': 0.,
            't1': 1.,
            'x0': 1,
            'num': 100,
        }

        self.model_args = []
        self.solver_options = {
            't_eval': np.linspace(
                self.ivp_options['t0'],
                self.ivp_options['t1'],
                num=self.ivp_options['num']
                )
        }

        self.events = {
                'default': self.flatlining_default,
                'dynamic': self.flatlining_dynamic,
            }

    def update_event_func(self):
        """
        Update the event function and ensures that solver_options are set.

        Raises:
            Exception: \\n
             Warning: Event handling got inadequadt event function,
             falling back to default

        """

        try:
            kw = self.solver_options['events']
            self.solver_options['events'] = self.events[kw]

        except Exception:
            print(
                """
                Warning: Event handling got inadequadt event function,
                falling back to default
                """
                )
            self.solver_options['events'] = self.events['default']

    def flatlining_default(self, t, x_0, *args):
        """
        The default flatlining function for determining the terminal event of
        the dynamical system.

        Based on the event handling schemes of scipy.integrate.solve_ivp
        (1.7.3), has the same call signature as calc_update_stimuli(). The
        method checks whether the system's Lyapunov function
        has decreased beneath a certain threshold.

        Args:
            t (float):\n
             Current time step in numeric ODE evaluation \n
            x_0 (array):\n
             Current state vector of the DS. \n
            args (iterable):\n
             Model specific tuple of parameters, same signature as needed for
             calc_update_stimuli. \n

        Returns:
            float: \n
             z, the signature of sign change, signaling the internal
             threshold being passed.

        """

        F, dF = self.calc_cost_stimuli(t, x_0, *args)
        dF_abs = np.linalg.norm(dF)
        quality = np.round(np.divide(dF_abs, F), self.null_decimal)

        z = quality - np.power(10., -(self.null_decimal-1))

        # print(f'ref: {z}')
        # print(f't: {t}')
        return z

    def flatlining_dynamic(self, t, x_0, *args):

        """
        The dynamic flatlining function for determining the terminal event of
        the dynamical system.

        Based on the event handling schemes of scipy.integrate.solve_ivp
        (1.7.3), has the same call signature as calc_update_stimuli(). The
        method checks whether the magnitude of relative change of the
        state vector x_0 has decreased beneath a certain threshold.

        Args:
            t (float):\n
             Current time step in numeric ODE evaluation \n
            x_0 (array):\n
             Current state vector of the DS. \n
            args (iterable):\n
             Model specific tuple of parameters, same signature as needed for
             calc_update_stimuli. \n

        Returns:
            float: \n
             z, the signature of sign change, signaling the internal
             threshold being passed.

        """

        dx = self.calc_update_stimuli(t, x_0, *args)
        dx_abs = np.absolute(dx)
        rel_r = np.divide(dx_abs, x_0)
        quality = np.round(np.linalg.norm(rel_r), self.null_decimal)

        z = quality-np.power(10., -(self.null_decimal-1))

        return z

    for f in [flatlining_default, flatlining_dynamic]:
        f.terminal = True
        f.direction = 1.

    def set_model_parameters(self, model_pars):
        """
        Set internal model arguments array.

        Args:
             model_pars (dict):\n
              Model argument dictionary generated on program initilization. \n

        """

        for k, v in model_pars.items():

            self.model_args.append(v)

    def set_solver_options(self, solv_opt):
        """
        Set internal solver_options and update event function.

        Args:
             solv_opt (dict):\n
              Solver argument dictionary generated on program initilization. \n

        """

        for k, v in solv_opt.items():

            self.solver_options[k] = v

        self.update_event_func()

    def calc_update_stimuli(self, t, x_0, *args):
        """
        The dynamic system's temporal update rule, computing the gradient -dF
        for dx/dt.

        Args:
            t (float):\n
             Current time step in numeric ODE evaluation \n
            x_0 (array):\n
             Current state vector of the DS. \n
            args (iterable):\n
             Model specific tuple of parameters, needed to evaluate stimulus
             functions \n

        Returns:
            array: \n
             The computed change dx to the state vector x_0.

        """

        dx = np.array()

        return dx

    def calc_cost_stimuli(self, t, x_0, *args):
        """
        Computes the dynamic system's Lyapunov function and its gradient.

        Args:
            t (float):\n
             Current time step in numeric ODE evaluation \n
            x_0 (array):\n
             Current state vector of the DS. \n
            args (iterable):\n
             Model specific tuple of parameters, needed to evaluate stimulus ]
             functions \n

        Returns:
            iterable: \n
             The computed gradient dF to the current cost function F, as well
             as F itself.

        """

        F = np.array()
        dF = np.array()

        return F, dF



# @Author: Felix Kramer <kramer>
# @Date:   23-06-2021
# @Email:  kramer@mpi-cbg.de
# @Project: phd_network_remodelling
# @Last modified by:   felix
# @Last modified time: 2022-07-02T13:15:02+02:00

import numpy as np
from dataclasses import dataclass, field
# custom
from .base import model


@dataclass
class murray(model):
    """
    The network adaptation model according to Murray, based on flow
    dissipation-volume minimization as published in:
     Murray, The Physiological Principle of Minimum Work, PNAS, 1926

    The system's cost function is given as:\n
    .. math::
        \\Gamma = \\sum_{e}\\left(\\frac{f_e^2}{C_e} + ar_e^2\\right)
    For minimizing this cost one iteratively computes the flow f_e,
    pressure \\Delta p_e landscapes. The system's derived gradient descent
    dynamics are given as:\n
    .. math::
        \\frac{dr_e}{dt} \\propto \\left[\\alpha_1\\left(
        \\frac{\\Delta p_er_e}{L_e}\\right)^2-\\alpha_0\\right] r_e
    Attributes:
        pars (dict):\n
         The specific model parameters \\alpha_0, \\alpha_1. \n

        ivp_options (dict):\n
         Information to generate the internal solver_options. Providing t0,t1
         number of evaluation and x0.\n
        model_args (list):\n
         Model specific paramerets need to evaluate the update rules of the DS
         \n
        solver_options (dict):\n
         Specifying runtime and evaluation marks. \n
        events (dict):\n
         Events to consier, here in gernal flatlining events which allow for
         early termination of stiff simulations. \n
        null_decimals (in):\n
         description \n

    Methods:
        init()
            The model post_init function.
        update_event_func()
            Update the event function and ensures that solver_options are set.
        flatlining_default(t, x_0, *args)
            The default flatlining function for determining the terminal event
            of the dynamical system.
        flatlining_dynamic(t, x_0, *args)
            The dynamic flatlining function for determining the terminal event
            of the dynamical system.
        set_model_parameters(model_pars)
            Set internal model arguments array.
        set_solver_options(solv_opt)
            Set internal solver_options and update event function.
        calc_update_stimuli(t, x_0, *args)
            The dynamic system's temporal update rule, computing the gradient
            -dF for dx/dt.
        calc_cost_stimuli(t, x_0, *args)
            Computes the dynamic system's Lyapunov function and its gradient.
        get_stimuli_pars(self, flow, x_0)
            Update flow & pressure landscapes, recompute conductivity as well
            as squared quantities.

    """

    pars: dict = field(default_factory=dict, init=True, repr=True)

    def __post_init__(self):

        self.init()

        self.model_args = [1., 1.]
        self.solver_options.update({'events': 'default'})
        if self.pars:
            self.set_model_parameters(self.pars)

    def set_model_parameters(self, model_pars):

        for k, v in model_pars.items():

            if 'alpha_0' == k:
                self.model_args[0] = v

            if 'alpha_1' == k:
                self.model_args[1] = v

    def set_solver_options(self, solv_opt):

        for k, v in solv_opt.items():

            self.solver_options[k] = v

        self.update_event_func()

    def calc_update_stimuli(self, t, x_0, flow, alpha_0, alpha_1):

        x_sq, p_sq = self.get_stimuli_pars(flow, x_0)

        s1 = 2.*alpha_1*np.multiply(p_sq, x_sq)
        s2 = alpha_0*np.ones(len(x_0))
        ds = np.subtract(s1, s2)
        dx = 2*np.multiply(ds, x_0)

        return dx

    def calc_cost_stimuli(self, t, x_0, flow, alpha_0, alpha_1):

        x_sq, p_sq = self.get_stimuli_pars(flow, x_0)

        f1 = alpha_1*np.multiply(p_sq, np.power(x_sq, 2))
        f2 = alpha_0 * x_sq
        F = np.sum(np.add(f1, f2))

        dF = -self.calc_update_stimuli(t, x_0, flow, alpha_0, alpha_1)

        return F, dF

    def get_stimuli_pars(self, flow, x_0):
        """
        Update flow & pressure landscapes, recompute conductivity as well as
        squared quantities.

        Invokes the internal flow circuit and recalcs flows and pressures as
        well as conductivities according to Hagen-Poiseuille's law of flows in
        cylindrical pipes.

        Args:
            flow (flow):\n
                A hailhydro flow object, describing the current flow landscape
                in the transport network.\n
            x_0 (array):\n
                Current state vector of the dynmic system.\n

        Returns:
            iterable: \n
                The squared state vector as well as the squared pressure
                landscape x_sq, p_sq

        """

        k = flow.circuit.scales['conductance']
        src = flow.circuit.nodes['source']

        x_sq = np.power(x_0, 2)
        conductivity = flow.calc_conductivity_from_cross_section(x_sq, k)
        p_sq, q_sq = flow.calc_sq_flow(conductivity, src)

        return x_sq, p_sq



# @Author: Felix Kramer <kramer>
# @Date:   23-06-2021
# @Email:  kramer@mpi-cbg.de
# @Project: phd_network_remodelling
# @Last modified by:   felix
# @Last modified time: 2022-07-02T14:16:15+02:00

import numpy as np
import copy
from dataclasses import dataclass, field
# custom
from hailhydro.flow_random import FlowRandom as FlowRandom
from .base import model


def dualFlowRandom(dualCircuit, *args):

    dualFlow = []
    for i, cs in enumerate(dualCircuit.layer):

        new_args = copy.deepcopy(args)
        for arg in new_args:
            for k, v in arg.items():
                arg[k] = v[i]

        dualFlow.append(FlowRandom(cs, *new_args))
        dualFlow[-1].e_adj = dualCircuit.e_adj[:]
        dualFlow[-1].dist_adj = dualCircuit.dist_adj[:]

    return dualFlow


@dataclass
class kramer(model):
    """
    The network adaptation model for for two entangled networks, according
    to Kramer et al, based on minimization of noisy disspation-volume in
    combination with intertwinedness retrictions/couplings:
        Kramer and Modes, How to pare a pair: Topology control and pruning in
        intertwined complex networks, PRR, 2020

    The system's cost function is given as:\n
    .. math::
        \\Gamma = \\sum_{e} \\left( C_e \\langle \\Delta p_e^2 \\rangle
        + a r_e^{2} \\right) + b \\sum_{ee'} \\Delta r_{ee'}^{\\varepsilon}
    Attributes:
        pars (dict):\n
         The specific model parameters p0 (growth rate), p1 (coupling), p2
         (volume penalty), p3 (fluctuation) and coupling exponent \\varepsilon.

        ivp_options (dict):\n
         Information to generate the internal solver_options. Providing t0,t1
         number of evaluation and x0.\n
        model_args (list):\n
         Model specific paramerets need to evaluate the update rules of the DS
         \n
        solver_options (dict):\n
         Specifying runtime and evaluation marks. \n
        events (dict):\n
         Events to consier, here in gernal flatlining events which allow for
         early termination of stiff simulations. \n
        null_decimals (in):\n
         description \n

    Methods:
        init()
            The model post_init function.
        update_event_func()
            Update the event function and ensures that solver_options are set.
        flatlining_default(t, x_0, *args)
            The default flatlining function for determining the terminal event
            of the dynamical system.
        flatlining_dynamic(t, x_0, *args)
            The dynamic flatlining function for determining the terminal event
            of the dynamical system.
        set_model_parameters(model_pars)
            Set internal model arguments array.
        set_solver_options(solv_opt)
            Set internal solver_options and update event function.
        calc_update_stimuli(t, x_0, *args)
            The dynamic system's temporal update rule, computing the gradient
            -dF for dx/dt.
        calc_cost_stimuli(t, x_0, *args)
            Computes the dynamic system's Lyapunov function and its gradient.
        get_stimuli_pars(self, flow, x_0)
            Update flow & pressure landscapes, recompute conductivity as well
            as squared quantities.

    """

    pars: dict = field(default_factory=dict, init=True, repr=True)

    def __post_init__(self):

        self.init()

        self.solver_options.update({'events': 'dynamic'})
        if self.pars:
            self.set_model_parameters(self.pars)

    def update_event_func(self):

        self.solver_options['events'] = self.events[
                                            self.solver_options['events']
                                            ]

    def set_solver_options(self, solv_opt):

        for k, v in solv_opt.items():

            self.solver_options[k] = v

        self.update_event_func()

    def calc_update_stimuli(self, t, x_0, flow, p_0, p_1, p_2, p_3, coupling):

        # pruning
        sgl = np.where(x_0 <= 0.)[0]
        x_0[sgl] = np.power(10., -10)
        # x_0[sgl]

        idxSets = [len(f.circuit.edges['label']) for f in flow]
        sgn = coupling / np.absolute(coupling)

        dx_pre = []
        x_sep = [x_0[:idxSets[0]], x_0[idxSets[0]:]]

        for i, idx in enumerate(idxSets):

            f = flow[i]
            x = x_sep[i]
            f.set_effective_source_matrix(p_3[i])

            # calc flows
            x_sq, p_sq = self.get_stimuli_pars(f, x)
            x_cb = np.multiply(x_sq, x)

            # calc interaction
            cpl = np.zeros(idx)
            for j, e in enumerate(f.e_adj):

                dr = 1. - (x_sep[0][e[0]] + x_sep[1][e[1]])
                force = sgn * (dr**coupling)
                cpl[e[i]] += p_1[i] * force

            # calc total feedback
            shear_sq = np.multiply(p_sq, x_cb)
            vol = p_2[i] * x
            diff_shearvol = np.subtract(shear_sq, vol)

            dx_pre.append(p_0[i]*np.add(diff_shearvol, cpl))

        dx = np.concatenate((dx_pre[0], dx_pre[1]))

        # pruning
        dx[sgl] = 0.

        return dx

    def get_stimuli_pars(self, flow, x_0):

        k = flow.circuit.scales['conductance']

        x_sq = np.power(x_0, 2)
        conductivity = flow.calc_conductivity_from_cross_section(x_sq, k)
        p_sq, q_sq = flow.calc_sq_flow_effective(conductivity)

        return x_sq, p_sq

    def prune(self, t, x_0, flow, p_0, p_1, p_2, p_3, coupling):
        """
        Check whether a vessel collapsed, i.e. negative radii appear during
        integration. If so, handle it by pruning the vessel.

        Args:
            t (float):\n
             Current time step in numeric ODE evaluation \n
            x_0 (array):\n
             Current state vector of the DS. \n
            args (iterable):\n
             Model specific tuple of parameters, needed to evaluate stimulus ]
             functions \n

        Returns:
            int:
                1: do not prunr and keep updating
                0: vessel collapsed, stop updating

        """
        f = 1
        if np.any(x_0 < 0):
            f = 0

        return f



# @Author: Felix Kramer <felix>
# @Date:   2022-06-28T16:24:41+02:00
# @Email:  felixuwekramer@proton.me
# @Filename: binder.py
# @Last modified by:   felix
# @Last modified time: 2022-07-01T15:12:56+02:00

from . import murray, bohn, corson, meigel, kramer
from hailhydro import flow_init, flow_random, flux_overflow

modelBinder = {
    'default': murray.murray,
    'murray': murray.murray,
    'bohn': bohn.bohn,
    'corson': corson.corson,
    'meigel': meigel.meigel,
    'link': meigel.link,
    'kramer': kramer.kramer,
    # 'volume': meigel.volume,
}
circuitBinder = {
    'murray': flow_init.Flow,
    'bohn': flow_init.Flow,
    'corson': flow_random.FlowRandom,
    'meigel': flux_overflow.Overflow,
    'link': flux_overflow.Overflow,
    'kramer': kramer.dualFlowRandom,
    # 'volume': flux_overflow.Overflow,
}
